from import_all_defs import import_all_defs


def test_load_all_defs():
    defs = import_all_defs(load_defs=True)
    assert len(defs) > 0
    assert type(defs) == dict
    # Test that functions were loaded
    funcs = defs.get("functions")
    assert funcs
    for key, definition in funcs.items():
        assert hasattr(definition, "__call__")


def test_find_all_defs():
    defs = import_all_defs(load_defs=False)
    assert len(defs) > 0
    assert type(defs) == dict
    # Test that definition paths were returned
    funcs = defs.get("functions")
    assert funcs
    for key, definition in funcs.items():
        assert type(definition) == str



"""
Import all definitions under this directory, so you can import as "from <directory name> import <definition name>"
Definitions define interfaces that may be implemented in various ways over time.
Definitions may be any type of Python variable or value: functions, classes, objects, constants, etc.

Since modules may import other modules, only a reference to each definition
file is imported at first and then each module is loaded when first used.
"""
import importlib
import ntpath
import os
import sys
import types
from importlib.util import module_from_spec, spec_from_file_location
from os.path import join

import git

space_path = os.path.dirname(__file__)
space_name = os.path.basename(space_path)
reload_defs = os.environ.get("RELOAD_DEFs", False)


def _import_from_path(path: str):
    """Given a filepath, return the module"""
    name = os.path.basename(path).replace(".py", "")
    mod = None
    try:
        spec = spec_from_file_location(name, path)
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)
    except ModuleNotFoundError as e:
        if not str(e).endswith(f" {name}"):
            # Failed import of some other module in the module file
            raise e
    return mod


def _def_from_path(path: str):
    """Return definition function or object at provided path"""
    mod = _import_from_path(path)
    if mod:
        try:
            definition = getattr(mod, mod.__name__)
        except AttributeError:
            return None
    else:
        return None
    return definition


def import_space_defs(space_path, load_defs: bool = True, search_def: str = None):
    """Import all modules in the given directory path and return a dictionary of modules
    keyed by definition name
    if search_def is set, will stop search after finding the provided module name"""
    space_defs = {}
    # Find directories at the space level
    for dirpath, dirs, files in os.walk(space_path):
        dirsplits = dirpath.split(os.sep)
        # For each directory, find the definition file
        for dir_name in dirs:
            if dir_name.startswith("_"):
                # e.g. __pycache__
                continue
            def_filename = join(dirpath, dir_name, dir_name + ".py")
            try:
                if load_defs:
                    # Get function with same name inside module
                    definition = _def_from_path(def_filename)
                else:
                    definition = def_filename
                if definition:
                    space_defs[dir_name] = definition
                if search_def is not None and search_def == dir_name:
                    break
            except FileNotFoundError:
                pass
    return space_defs


class ReloadWrapper(object):
    """If reload_def, execute the function's defining code on every import.
    Otherwise, load only once"""

    def __init__(self, def_filename, space_name, mod_name, reload_def=False):
        """save the file location of the definition to load later"""
        self.def_filename = def_filename
        if not reload_def:
            self.space_name = space_name
            self.mod_name = mod_name
            self.load_definition = self.load_definition_once

    def __call__(self, *args, **kwargs):
        """Load the definition and call it with arguments"""
        if hasattr(self, '_definition'):
            # Once already loaded
            try:
                return self._definition(*args, **kwargs)
            except TypeError:
                # Not loaded yet
                pass
        self._definition = self.load_definition()
        try:
            return self._definition(*args, **kwargs)
        except TypeError as e:
            if self._definition is None:
                print(f"No definition found at {self.def_filename}", file=sys.stderr)
            else:
                raise e

    def load_definition(self):
        """Return loaded definition"""
        definition = _def_from_path(self.def_filename)
        self.call = definition
        return definition

    def load_definition_once(self):
        """since reload_def is false, only load the first time
        and change __call__ to be the loaded function itself.
        """
        definition = _def_from_path(self.def_filename)
        sys.modules[self.space_name].__dict__[self.mod_name] = definition
        sys.modules[f"{self.space_name}.{self.mod_name}"] = definition
        self.call = definition
        return definition

    def help(self):
        """Return information about the definition"""
        return help(self._definition)

    def definition(self):
        if self._definition:
            return self._definition
        self._definition = self.load_definition()
        return self._definition

    def __getattribute__(self, *args, **kwargs):
        try:
            return super().__getattribute__(*args, **kwargs)
        except AttributeError:
            return self.definition.__getattribute__(*args, **kwargs)

    def __setattr__(self, *args, **kwargs):
        try:
            return super().__setattr__(*args, **kwargs)
        except AttributeError:
            return self.definition.__setattr__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        try:
            return super().__getitem__(*args, **kwargs)
        except AttributeError:
            return self.definition.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        try:
            return super().__setitem__(*args, **kwargs)
        except AttributeError:
            return self.definition.__setitem__(*args, **kwargs)

    def __str__(self):
        try:
            return str(self._definition)
        except Exception:
            return str(self.definition)

    def __repr__(self):
        try:
            return repr(self._definition)
        except Exception:
            return repr(self.definition)


def import_space(space_path, reload_defs=False, wrapper=ReloadWrapper):
    """Under the directory space_path,
        find all modules and attach to sys.modules for importing
    since modules may import other modules that need to be imported first,
        we always find and wrap all modules in ReloadWrapper
        even if reload_defs is False.
    If reload_defs is True, rerun the module's code every time it's imported.
    If reload_defs
    """
    # Get file names of definitions. Don't load them until they're needed.
    space_defs = import_space_defs(space_path, load_defs=False)
    # Create module for space and attach all its definitions to it
    # Also set each definition in sys.modules
    space_name = os.path.basename(space_path)
    space_mod = importlib.import_module(".", package=space_name)
    # space_mod = _import_from_path(space_path)
    if space_mod:  # space folder may be empty
        space_mod.defs = space_defs
        for def_name, def_filename in space_defs.items():
            mod_def = wrapper(
                def_filename,
                space_name=space_name,
                mod_name=def_name,
                reload_def=reload_defs,
            )
            space_mod.__dict__[def_name] = mod_def
            sys.modules[f"{space_name}.{def_name}"] = mod_def
        sys.modules[space_name] = space_mod
        # set space as attribute of current module
        globals()[space_name] = space_mod


wrapper = ReloadWrapper
import_space(space_path, reload_defs, wrapper)
# Note: Deleting the below variables causes problems if importing multiple spaces at once
# using the __init__.py file at the repository root.

try:
    del ReloadWrapper
    del git
    del import_space
    del import_space_defs
    del importlib
    del join
    del ntpath
    del path
    del reload_defs
    del space_name
    del space_path
    del src_init_path
    del types
    del wrapper

    del defs
    del objects
except Exception:
    pass


"""
Module provides a simple cubic_rectification function.
"""

import numpy as np


def cubic_rectification(x, verbose=False):
    """
    Rectified cube of an array.

    Parameters
    ----------
    x : numpy.ndarray
        Input array.
    verbose : bool, optional
        Whether to print out details. Default is ``False``.

    Returns
    -------
    numpy.ndarray
        Elementwise, the cube of `x` where it is positive and ``0`` otherwise.

    Note
    ----
    This is a sample function, using a numpy docstring format.

    Note
    ----
    The use of intersphinx will cause :class:`numpy.ndarray` to link to
    the numpy documentation page.
    """
    if verbose:
        print("Cubing and then rectifying {}".format(x))
    return np.maximum(0, x ** 3)



# apt-get install libffi-dev # To avoid installation error: ModuleNotFoundError: No module named '_ctypes'
from repo_utils import install_dependencies
import os

# Install requirements of all definitions in the repository
install_dependencies(os.path.dirname(__file__))


from repo_utils import def_from_dir, find_definition


def import_definition(name: str, namespace: str = None):
    """Import a definition based on its name without its path
    Returns the imported function or object, not the module
    namespace should be a directory or path from <repo_root>/src
    If the module name is not found, return None"""
    def_dir = find_definition(name, namespace=namespace)
    if def_dir:
        definition = def_from_dir(def_dir)
        if definition is None:
            print(f'Definition {def_dir} not found. Does the Python filename have the same name as the folder name?')
        return definition
    else:
        print(f'Definition {name} not found.')


import os
import pathlib
import shutil

from define import define

curr_dir = pathlib.Path(__file__).parent.absolute()


def test_define():
    name = "test_def_test"
    name_dir = os.path.join(curr_dir, "test_def_test")
    # Remove test directory if it's already there
    try:
        shutil.rmtree(name_dir)
    except Exception:
        pass

    define(name_dir)
    # Folder
    assert os.path.exists(name_dir)
    # Definition file
    assert os.path.exists(os.path.join(name_dir, name + ".py"))
    # Test file
    assert os.path.exists(os.path.join(name_dir, "test_" + name + ".py"))

    # Remove test directory
    shutil.rmtree(name_dir)


def test_define_path():
    name_dir = os.path.join(curr_dir, "test_def_test")
    # Remove test directory if it's already there
    try:
        shutil.rmtree(name_dir)
    except Exception:
        pass

    define(name_dir)
    # Folder
    assert os.path.exists(name_dir)
    # Definition file
    assert os.path.exists(os.path.join(name_dir, "test_def_test.py"))
    # Test file
    assert os.path.exists(os.path.join(name_dir, "test_test_def_test.py"))

    # Remove test directory
    shutil.rmtree(name_dir)



import pytest


def pytest_on_path(path: str, test_func: str = None):
    """Given a filepath, run pytest"""
    if test_func:
        path = f"{path}::{test_func}"
    print("pytest_on_path", "path", path)
    return pytest.main([path])



from os.path import exists, join

from repo_utils import find_definition, list_dependencies


def enable_installing_dependencies(path_or_def: str = None, package_name: str = None):
    """
    Create an __init__.py file in the necessary folders so that these are included in the package installation.
    Optionally specify a path from which to determine dependencies,
        so only the required definitions will be installed.

    Args:
        - path_or_def (str): If a string with no /, locate the definition with find_definition
            and identify all dependencies by imports recursively.
            If a directory path, search for imports among all files at that path recursively in subfolders.
            If not provided, the repository root will be used as the path and an init file
                will be created in every directory that has a file with the same name.
        - package_name (str): The package name used when importing (only these imports will be listed).

    Returns:
        - init_filepaths (list of str): Filepaths of init files created.
    """
    init_filepaths = []
    deps = list_dependencies(path_or_def, package_name)
    for dep in deps:
        def_path = find_definition(dep)
        if def_path is None:
            continue
        init_filepath = join(def_path, "__init__.py")
        if not exists(init_filepath):
            with open(init_filepath, "w") as file:
                pass
            init_filepaths.append(init_filepath)
    return init_filepaths



import sys
from os import remove
from os.path import exists, join

from repo_utils import find_definition, pytest_on_path


def test_def(def_name: str, test_func: str = None, namespace: str = None):
    """
    Given the name of a definition to test_def,
    run the chosen test_def module on it.

    Args:
        - def_name (str): Definition name at which to find the test module file (test_<def_name>.py).
        - test_func (str): Test only this particular test function within the test module.

    """
    def_dir = find_definition(def_name, namespace=namespace)
    # In order to specify function, the specific
    # test_def file must be specified, so it
    # will only search within the test_def file named after
    # the definition
    if def_dir:
        init_filepath = join(def_dir, "__init__.py")
        if exists(init_filepath):
            remove(init_filepath)
        return pytest_on_path(join(def_dir), test_func=test_func)
    else:
        print(f"No definition found named {def_name}", file=sys.stderr)



import os
import re
from os.path import join, exists

from repo_utils import find_definition, get_repository_path


def list_dependencies(path_or_def: str, package_name: str = None):
    """
    Determine dependencies within a repository or package.

    Args:
        - path_or_def (str): If a string with no /, locate the definition with find_definition
            and identify all dependencies by imports recursively.
            If a directory path, search for imports among all files at that path recursively in subfolders.
        - package_name (str): The package name used when importing (only these imports will be listed).
            If not provided, will use the path set at __meta__.py at the repository root (TODO).

    Returns:
        - dependencies (list): List of unique dependent definitions.
    """
    if path_or_def is None:
        path_or_def = get_repository_path()
    # if package_name is None:
    # exec(read("TODO/__meta__.py"), meta)
    if package_name is None:
        raise TypeError("list_dependencies: You must specify package_name used for imports.")

    filenames_to_search = []
    dependencies = []
    if not exists(path_or_def):
        # path_or_def is a definition name.
        # Search recursively for dependencies
        sub_deps = [path_or_def]
        already_checked = set()
        while len(sub_deps):
            dep = sub_deps.pop()
            if dep in already_checked:
                continue
            def_folder = find_definition(dep)
            if def_folder is None:
                print(f'list_dependencies: Definition {dep} not found.')
                continue
            def_path = join(find_definition(dep), f"{dep}.py")
            file_dependencies = list_file_dependencies(def_path, package_name)
            already_checked.add(dep)
            sub_deps.extend(file_dependencies)
            dependencies.extend(file_dependencies)
    else:
        # path_or_def is a folder path.
        for dirpath, dirs, files in os.walk(path_or_def):
            for filename in files:
                if not filename.endswith(".py"):
                    continue
                filepath = join(dirpath, filename)
                file_dependencies = list_file_dependencies(filepath, package_name)
                dependencies.extend(file_dependencies)
    return list(set(dependencies))


def list_file_dependencies(filepath, package_name):
    """
    Given a filepath, return a list of dependencies based on imports in just that file.

    Returns:
        - dependencies (list): List of dependent definitions in the order in which they were found.

    Implementation:
        1. Include the filename as a dependency if it's the same name as the containing folder.
    """
    dependencies = []
    split_filepath = filepath.split(os.sep)
    # 1. Include the filename as a dependency if it's the same name as the containing folder.
    if len(split_filepath) > 1:
        def_name = split_filepath[-1].strip().split(".")[0]
        if def_name == split_filepath[-2]:
            dependencies.append(split_filepath[-2])
    with open(filepath, "r") as file:
        within_parens = False
        after_slash = False
        for line in file.readlines():
            line = line.strip()
            if not after_slash and not within_parens and not line.startswith("from"):
                continue
            if after_slash or within_parens:
                after_slash = False
                if ')' in line:
                    within_parens = False
                    line = line.replace(')', ' ')
                split_imported = line.split(",")
                for imported in split_imported:
                    imported = imported.strip()
                    if imported:
                        dependencies.append(imported)
            line = line[4:].strip()  # Remove from.
            if line.startswith(package_name) or after_slash or within_parens:
                split_line = line.split("import")
                if len(split_line) > 1:
                    imported = split_line[1].strip()
                    if imported.startswith('('):
                        within_parens = True
                        imported = imported[1:]
                    elif imported.endswith('\\'):
                        after_slash = True
                    split_imported = imported.split(",")
                    for imported in split_imported:
                        imported = imported.strip()
                        if imported:
                            dependencies.append(imported)
    return dependencies



from os.path import join

from list_dependencies import list_dependencies

from repo_utils import find_definition, get_repository_path


def test_list_def_dependencies():
    deps = list_dependencies("find_definition", "repo_utils")
    assert deps
    assert "find_definition" in deps


def test_parentheses():
    deps = list_dependencies("simulate_definition", "repo_utils")
    assert deps
    assert "find_definition" in deps
    assert "import_from_dir" in deps
    assert "import_definition" in deps


def test_list_recursive_dependencies():
    deps = list_dependencies("enable_installing_dependencies", "repo_utils")
    assert deps
    expected_deps = [
        "find_definition",
        "get_repository_path",
        "list_dependencies",
        "enable_installing_dependencies",
    ]
    for dep in expected_deps:
        assert dep in deps


def test_list_folder_dependencies():
    deps = list_dependencies(find_definition("find_definition"), "repo_utils")
    assert deps
    assert "find_definition" in deps

    deps = list_dependencies(
        join(get_repository_path(), "repo_utils", "project"), "repo_utils"
    )
    assert deps
    expected_deps = [
        "find_definition",
        "get_repository_path",
        "install_def",
        "list_dependencies",
        "start_command_line",
        "test_def",
        "import_from_dir"
    ]
    for dep in expected_deps:
        assert dep in deps


def test_dep_from_import():
    deps = list_dependencies(find_definition("install_def"), "repo_utils")
    assert deps
    assert "find_definition" in deps
    assert "install_def" in deps


def test_multi_imports_on_same_line():
    deps = list_dependencies(find_definition("test_def"), "repo_utils")
    assert deps
    assert "find_definition" in deps
    assert "pytest_on_path" in deps



from os.path import join

from repo_utils import find_definition, get_repository_path, import_from_dir


def test_import_same_dir():
    mod = import_from_dir(join(get_repository_path(), "src", "functions"))
    assert mod



import errno
import ntpath
import os
from os.path import join


def define(name: str):
    """Create definition files for definition at name
    name can be a path"""
    # Create folder
    path = os.path.abspath(os.path.expanduser(name))
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise e
    name = ntpath.basename(name)

    # Create definition file
    if "/objects/" in path:
        def_contents = f"""
class {name}():
    \"\"\"  \"\"\"
    pass
"""
    else:
        def_contents = f"""
def {name}():
    \"\"\"

    Args:
        - 

    Returns:
        - 
    \"\"\"
    pass
"""
    def_filename = join(path, name + ".py")
    with open(def_filename, "w") as file:
        file.write(def_contents)

    # Create test file
    test_contents = f"""from {name} import {name}

def test_{name}():
    pass
"""
    test_filename = join(path, "test_{}.py".format(name))
    with open(test_filename, "w") as file:
        file.write(test_contents)


def snake_to_camel(text):
    """Convert from this_style to ThisStyle"""
    text = "".join(word.capitalize() or "_" for word in text.split("_"))
    return text



import os
from os.path import join
from pathlib import Path
from shutil import move

from repo_utils import find_definition


def rename_def(def_name: str, target_name: str, namespace: str = None):
    """
    Change the name of a definition
    Make any needed changes to its folder, files,
    and any imports of the definition
    The definition with name def_name in namespace will be
    changed to target_name
    Return True if successful, False otherwise (e.g. no definition found)
    """
    def_dir = find_definition(def_name, namespace)
    if def_dir:
        path = Path(def_dir)
        parent_dir = path.parent
        target_dir = os.path.join(parent_dir, target_name)
        if os.path.exists(target_dir):
            print(f"Aborting move since directory exists: {target_dir}")
            return False
        move(def_dir, target_dir)
        # Use new name on all files of the directory
        # Note: This doesn't search files recursively
        # and doesn't include directories
        files = os.listdir(target_dir)
        for filename in files:
            # Replace file names
            source_path = join(target_dir, filename)
            if def_name in filename and os.path.isfile(source_path):
                new_name = filename.replace(def_name, target_name)
                move(source_path, join(target_dir, new_name))
                filename = join(target_dir, new_name)
            # Replace mentions in files
            if os.path.isfile(filename):
                new_text = None
                with open(filename, "r") as file:
                    text = file.read()
                    if text:
                        new_text = text.replace(def_name, target_name)
                if new_text:
                    with open(filename, "w") as file:
                        file.write(new_text)
        return True
    print(f'Definition {def_dir} not found.')
    return False



from repo_utils import list_dependencies, install_def

def install_dependencies(path_or_def: str, package_name: str, run_install_scripts=True):
    """

    Args:
        - path_or_def (str): If a string with no /, locate the definition with find_definition
            and identify all dependencies by imports recursively.
            If a directory path, search for imports among all files at that path recursively in subfolders.
        - package_name (str): The package name used when importing (only these imports will be listed).
            If not provided, will use the path set at __meta__.py at the repository root (TODO).
        - run_install_scripts (bool): If install is False, install.sh and install.py won't be run.

    Returns:
        -
    """
    for dep in list_dependencies(path_or_def, package_name):
        install_def(dep, install=run_install_scripts)



# `name` is the name of the package as used for `pip install package`
name = "rc-repo-utils"
# `path` is the name of the package for `import package`
# path = name.lower().replace("-", "_").replace(" ", "_")
path = "repo_utils"
# Your version number should follow https://python.org/dev/peps/pep-0440 and
# https://semver.org
version = "0.1.dev5"
author = "Joshua Mathias"
author_email = "joshuaamathias@gmail.com"
description = "Repository utilities to facilitate collaborative and modularized coding."  # One-liner
url = ""  # your project homepage
license = "MIT"  # See https://choosealicense.com



import os
from os.path import join

import git

from repo_utils import def_from_dir


def import_all_defs(
    space_level: int = 1, load_defs: bool = True, search_def: str = None
):
    """Import all modules in the repository and return a dictionary of modules
    keyed by definition name
    if search_def is set, will stop search after finding the provided module name"""
    repo = git.Repo(os.getcwd(), search_parent_directories=True)
    repo_root = repo.working_tree_dir
    src_root = join(repo_root, "src")
    defs = {}
    # Find directories at the space level
    level_offset = len(src_root.split(os.sep)) - 1
    for dirpath, dirs, files in os.walk(src_root):
        dirsplits = dirpath.split(os.sep)
        level_count = len(dirsplits) - level_offset

        if level_count == space_level:
            # Use directories at this level as import spaces (namespaces)
            for dir_name in dirs:
                defs[dir_name] = {}
        elif level_count > space_level:
            if level_count == space_level + 1:
                # Set modules of space we're currently under
                # Note: os.walk is a depth-first search
                space_defs = defs.get(dirsplits[-1])
            # space_name = ntpath.basename(dirpath)
            # space_defs = defs[dir_name]
            # Use the
            # For each directory, find the definition file
            for dir_name in dirs:
                if dir_name.startswith("_"):
                    # e.g. __pycache__
                    continue
                def_path = join(dirpath, dir_name)
                print("def_path", def_path)
                try:
                    if load_defs:
                        # Get function with same name inside module
                        definition = def_from_dir(def_path)
                    else:
                        definition = def_path
                    print("definition", definition)
                    if definition:
                        space_defs[dir_name] = definition
                    if search_def is not None and search_def == dir_name:
                        break
                except FileNotFoundError:
                    pass
    return defs



from importlib.util import module_from_spec, spec_from_file_location
from os.path import basename, join


def import_from_dir(path: str):
    """Given a path to a directory with an __init__.py file, return the module"""
    name = basename(path)
    mod = None
    try:
        path = join(path, "__init__.py")
        spec = spec_from_file_location(name, path)
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)
        mod = getattr(mod, name)
    except (ModuleNotFoundError, AttributeError):
        pass
    return mod



import os
from os.path import join
import inspect
import git


def find_definition(def_name: str, containing_folder: str=None, namespace: str = None):
    """
    Searches among modules directories and
    returns the directory path that matches def_name
    under the specified namespace directory

    Args:
        - containing_folder (str): Search only within this directory path. Current working directory by default.
        - namespace (str): Within containing_folder, find a folder with this name and search only under this folder.
    """
    if containing_folder is None:
        previous_frame = inspect.currentframe().f_back
        previous_previous_frame = previous_frame.f_back
        (filename, line_number, function_name, lines, index) = inspect.getframeinfo(previous_previous_frame)
        if "<frozen" in filename or 'site-packages' in filename:
            (filename, line_number, function_name, lines, index) = inspect.getframeinfo(previous_frame)
        if 'repo_utils' in filename:
            # When using the repo_utils command line, use repository of the current working directory.
            filename = os.getcwd()
        repo = git.Repo(filename, search_parent_directories=True)
        src_root = repo.working_tree_dir
    else:
        src_root = containing_folder
    # src_root = join(repo_root, repo.name) # subfolder with same name
    # level_offset = len(src_root.split(os.sep)) - 1

    if namespace is None:
        for dir_path, dirs, _ in os.walk(src_root):
            if dir_path == "build" or dir_path.endswith(os.sep + "build"):
                continue
            for dir_name in dirs:
                if dir_name == def_name:
                    return join(dir_path, dir_name)
    else:
        for dir_path, dirs, _ in os.walk(src_root):
            for dir_name in dirs:
                if dir_name == "build":
                    continue
                if not namespace or dir_name == namespace:
                    for space_path, space_dirs, space_files in os.walk(
                        join(dir_path, dir_name)
                    ):
                        for def_dir in space_dirs:
                            if def_dir == def_name:
                                return join(space_path, def_dir)
                    if namespace:
                        break
    return None



import inspect
import os
from importlib.util import module_from_spec, spec_from_file_location
from os.path import basename, dirname, join


def import_from_path(path: str, root_dir: str = None):
    """Given a filepath, return the module

    Args:
        - path (str): Relative or absolute path to the module.
        - root_dir (str): Directory from which to start the relative path (e.g. __file__).

    Returns:
        - mod (ModuleType): The loaded module.
    """
    if not path.startswith(os.sep):
        if root_dir:
            path = join(dirname(root_dir), path)
        else:
            # Apply relative path from caller's directory.
            path = join(dirname(inspect.stack()[1].filename), path)
    name = basename(path).replace(".py", "")
    if not path.endswith(".py"):
        path += ".py"
    mod = None
    try:
        spec = spec_from_file_location(name, path)
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)
    except ModuleNotFoundError:
        pass
    return mod



""" Command line utility for repository """
import fire
import repo_utils


def start_command_line():
    """Command-line interface for the repository.
    Specify the definition to execute and then any arguments.
    e.g. "define <name>".
    The Fire library converts the specified function or object into a command-line utility.
    """

    fire.Fire(repo_utils)



import os
from os.path import exists, join

from enable_installing_dependencies import enable_installing_dependencies

from repo_utils import find_definition


def test_enable_installing_dependencies():
    test_path = None

    assert not exists(
        join(find_definition("enable_installing_dependencies"), "__init__.py")
    )
    init_filepaths = enable_installing_dependencies(
        "enable_installing_dependencies", package_name="repo_utils"
    )
    assert exists(
        join(find_definition("enable_installing_dependencies"), "__init__.py")
    )

    for filepath in init_filepaths:
        if exists(filepath):
            os.remove(filepath)



import os
import git
import inspect


def get_repository_path(starting_path=None):
    """Return the path of the root directory of this repository"""
    if starting_path is None:
        previous_frame = inspect.currentframe().f_back
        previous_previous_frame = previous_frame.f_back
        (starting_path, line_number, function_name, lines, index) = inspect.getframeinfo(previous_previous_frame)
        if "<frozen" in starting_path or 'site-packages' in starting_path:
            (starting_path, line_number, function_name, lines, index) = inspect.getframeinfo(previous_frame)
        if 'repo_utils' in starting_path:
            # When using the repo_utils command line, use repository of the current working directory.
            starting_path = os.getcwd()
    repo = git.Repo(starting_path, search_parent_directories=True)
    repo_root = repo.working_tree_dir
    return repo_root



import os
import pathlib
import shutil
from os.path import join

from rename_def import rename_def

from repo_utils import define

curr_dir = pathlib.Path(__file__).parent.absolute()


def test_rename_def():
    try:
        shutil.rmtree(join(curr_dir, "target_def"))
    except Exception:
        pass
    define(join(curr_dir, "source_def"))
    result = rename_def("source_def", "target_def", namespace="functions")
    assert result
    assert os.path.isdir(join(curr_dir, "target_def"))
    assert not os.path.isdir(join(curr_dir, "source_def"))
    assert os.path.isfile(join(curr_dir, "target_def", "target_def.py"))

    # Remove test directory
    try:
        shutil.rmtree(join(curr_dir, "target_def"))
    except Exception:
        pass



import types

import git
from import_from_path import import_from_path


def test_import_from_path():
    repo = git.Repo(__file__, search_parent_directories=True)
    repo_root = repo.working_tree_dir
    mod = import_from_path(
        f"{repo_root}/src/functions/project/definitions/define/define.py"
    )
    assert type(mod) == types.ModuleType


def test_import_relative():
    mod = import_from_path("import_from_path")
    assert type(mod) == types.ModuleType



from find_definition import find_definition


def test_find_definition():
    path = find_definition("define")
    assert "repo_utils" in path

    path = find_definition("TestObj")
    assert "repo_utils" in path

    path = find_definition("TestObj", namespace="functions")
    assert path is None

    path = find_definition("TestObj", containing_folder=find_definition('find_definition'))
    assert path is None


import unittest

from import_definition import import_definition


class TestImportDefinition(unittest.TestCase):
    def setUp(self):
        pass

    def test_import_definition(self):
        imported = import_definition("define")
        self.assertTrue(hasattr(imported, "__call__"))

        # Import definition that doesn't exist.
        imported = import_definition("definetest")
        self.assertEqual(imported, None)

    def test_scope(self):
        imported = import_definition("define", namespace="metrics")
        self.assertEqual(imported, None)

        # import an object
        imported = import_definition("TestObj")
        self.assertTrue(hasattr(imported, "__init__"))

    def test_import_attr(self):
        imported = import_definition("TestObj")
        assert imported.static_func() == "static"
        assert imported().func() == "func"

    def test_import_dict(self):
        imported = import_definition("TestDict")
        assert imported["item"] == "item"

    def test_str(self):
        imported = import_definition("define")
        assert str(imported).startswith("<function define at ")

        imported = import_definition("define")
        assert repr(imported).startswith("<function define at ")



import os

from repo_utils import import_from_path


def def_from_dir(path: str):
    """Return definition function or object at provided path"""
    if not path.endswith(".py"):
        path += os.sep + os.path.basename(path) + ".py"
    mod = import_from_path(path)
    try:
        definition = getattr(mod, mod.__name__)
    except AttributeError:
        return None
    return definition
